"""Pynecil - Python library to communicate with Pinecil V2 soldering irons via Bluetooth."""

from __future__ import annotations

import asyncio
import hashlib
import logging
import struct
from typing import TYPE_CHECKING, Any, cast

from bleak import BleakClient, BleakScanner
from bleak.backends.device import BLEDevice
from bleak.exc import BleakError

from . import const
from .exceptions import CommunicationError
from .types import (
    AnimationSpeed,
    AutostartMode,
    BatteryType,
    Characteristic,
    CharBulk,
    CharLive,
    CharSetting,
    DeviceInfoResponse,
    LanguageCode,
    LiveDataResponse,
    LockingMode,
    LogoDuration,
    OperatingMode,
    PowerSource,
    ScreenOrientationMode,
    ScrollSpeed,
    SettingsDataResponse,
    TempUnit,
)

if TYPE_CHECKING:
    from collections.abc import Callable

_LOGGER = logging.getLogger(__package__)


async def discover(timeout: float = 10) -> BLEDevice | None:
    """Discover Pinecil device.

    Parameters
    ----------
    timeout : float, optional
         Timeout to wait for detection before giving up, by default 10

    Returns
    -------
        BLEDevice | None
            The BLEDevice of a Pinecil or None if not detected
            before the timeout.

    """
    return await BleakScanner().find_device_by_filter(
        filterfunc=lambda _, advertisement: bool(
            str(const.SVC_UUID_BULK) in advertisement.service_uuids
        ),
        timeout=timeout,
    )


class Pynecil:
    """Client for Pinecil V2 Devices.

    Attributes
    ----------
    client_disconnected : bool
        Flag indicating whether the client has experienced an unexpected disconnect.
        Defaults to False.

    """

    client_disconnected: bool = False

    def __init__(
        self,
        address_or_ble_device: BLEDevice | str,
        disconnected_callback: Callable[[BleakClient], None] | None = None,
    ) -> None:
        """Initialize a Pynecil client.

        Parameters
        ----------
        address_or_ble_device : BLEDevice | str
            Bluetooth address of the Pinecil V2 device to connect to or BLEDevice object,
            received from a BleakScanner, representing it.
        disconnected_callback : Callable[[BleakClient], None] | None, optional
            Callback passed to BleakClient that will be scheduled in the event loop
            when the client is disconnected. The callable must take one argument,
            which will be the client object. Defaults to None.

        Notes
        -----
        If `address_or_ble_device` is a BLEDevice object, `device_info` will be initialized
        with its `name` and `address`. Otherwise, `device_info` will only have an `address`
        initialized.

        Callbacks for disconnection events can be specified via `disconnected_callback`.
        If not provided, a default callback will set `client_disconnected` to True upon
        disconnection.

        """
        if isinstance(address_or_ble_device, BLEDevice):
            self.device_info = DeviceInfoResponse(
                name=address_or_ble_device.name, address=address_or_ble_device.address
            )
        else:
            self.device_info = DeviceInfoResponse(address=address_or_ble_device)

        def _disconnected_callback(client: BleakClient) -> None:
            _LOGGER.debug("Disconnected from %s", client.address)
            self.client_disconnected = True
            self.device_info.is_synced = False

        self._client = BleakClient(
            address_or_ble_device,
            disconnected_callback=disconnected_callback or _disconnected_callback,
        )

    @property
    def is_connected(self) -> bool:
        """Check if the client is connected.

        Returns
        -------
        bool
            `True` if the client is connected, `False` otherwise.

        """
        return self._client.is_connected

    async def connect(self) -> None:
        """Establish or re-establish a connection to the device.

        This method ensures a connection to the device is established, handling both initial
        connections and reconnections if an unexpected disconnection occurred. If the device
        is not connected and there was a previous unexpected disconnect, it closes the stale
        connection before attempting to reconnect.

        Notes
        -----
        - If `self._client.is_connected` is `True`, the method returns without performing
        any connection operations.
        - If `self.client_disconnected` is `True`, indicating a previous unexpected
        disconnection, the method first closes the stale connection by calling `disconnect()`
        before attempting to establish a new connection.

        Raises
        ------
        BleakError
            If an error occurs during the connection attempt.

        """
        if not self._client.is_connected:
            if self.client_disconnected:  # close stale connection
                await self.disconnect()
            await self._client.connect()

    async def disconnect(self) -> None:
        """Disconnect from the Pinecil device."""
        await self._client.disconnect()
        self.client_disconnected = False

    async def get_device_info(self) -> DeviceInfoResponse:
        """Retrieve device information from the Pinecil V2 device.

        This method fetches details such as `name`, `address`, `build`, `device_sn`,
        and `device_id` of the connected Pinecil V2 device.

        Returns
        -------
        DeviceInfoResponse
            An object containing the retrieved device information.

        Raises
        ------
        CommunicationError
            If an error occurred while connecting and retrieving data from device.

        Notes
        -----
        If `self.device_info.is_synced` is `True`, it returns the cached `device_info`
        without performing a new data fetch.

        """
        if self.device_info.is_synced:
            return self.device_info

        try:
            self.device_info.build = await self.read(CharBulk.BUILD)
            self.device_info.device_sn = await self.read(CharBulk.DEVICE_SN)
            self.device_info.device_id = await self.read(CharBulk.DEVICE_ID)
            self.device_info.is_synced = True
        except (BleakError, TimeoutError) as e:
            _LOGGER.debug("Get device characteristics exception: %s", e)
            raise CommunicationError from e

        return self.device_info

    async def get_live_data(self) -> LiveDataResponse:
        """Fetch live sensor data from the device.

        Returns
        -------
        LiveDataResponse
            An object containing 14 sensor values retrieved from the
            bulk live data characteristic.


        Raises
        ------
        CommunicationError
            If an error occurred while connecting and retrieving data from device.

        """
        return await self.read(CharBulk.LIVE_DATA)

    async def get_settings(
        self, settings: list[CharSetting | int] | None = None
    ) -> SettingsDataResponse:
        """Fetch settings data from the device.

        Parameters
        ----------
        settings : list[CharSetting  |  int] | None, optional
            List of settings identified by a `CharSetting` item or their ID that should
            be retrieved from the device. Retrieves all settings if ommited.

        Returns
        -------
        SettingsDataResponse
            A dictionary containing lowercase names and normalized values of the settings
            retrieved from the device.

        Raises
        ------
        CommunicationError
            If an error occurred while connecting and retrieving data from device.

        Notes
        -----
        This method asynchronously reads data from the device for each specified setting,
        filtering based on the provided `settings` list.

        """
        if settings is None:
            settings = []
        tasks = [
            (characteristic.name.lower(), self.read(characteristic))
            for characteristic in CHAR_MAP
            if isinstance(characteristic, CharSetting)
            and (
                not settings
                or characteristic in settings
                or characteristic.value in settings
            )
        ]
        results = await asyncio.gather(*(task[1] for task in tasks))

        return cast(
            SettingsDataResponse,
            {key: value for (key, _), value in zip(tasks, results)},
        )

    async def read(self, characteristic: Characteristic) -> Any:
        """Read specified characteristic and decode the result.

        Parameters
        ----------
        characteristic : Characteristic
            The characteristic to retrieve from the device.

        Returns
        -------
        Any
            The value read from the characteristic and parsed with the corresponding decoder.

        Raises
        ------
        CommunicationError
            If an error occurred while connecting and retrieving data from device.

        """
        uuid, decode, *_ = CHAR_MAP.get(characteristic, (None, None))

        if not (decode and uuid):
            return None

        try:
            await self.connect()
            result = await self._client.read_gatt_char(uuid)
            _LOGGER.debug(
                "Read characteristic %s, result: %s", str(uuid), decode(result)
            )
        except (BleakError, TimeoutError) as e:
            _LOGGER.debug("Failed to read characteristic %s: %s", str(uuid), e)
            raise CommunicationError from e
        return decode(result)

    async def write(self, setting: CharSetting, value: Any) -> None:
        """Write to the specified characteristic.

        Parameters
        ----------
        setting : CharSetting
            The characteristic to write to.
        value : Any
            The value to write to the characteristic.

        Raises
        ------
        ValueError
            If no conversion or validation functions are found for the specified `setting`.
        CommunicationError
            If an error occurs while connecting to or writing data to the device.

        """
        uuid, _, convert, validate = CHAR_MAP.get(setting, (None, None, None, None))

        if not (convert and validate):
            raise ValueError(
                f"No conversion or validation functions found for {setting}"
            )

        data = validate(convert(value))
        try:
            await self.connect()
            await self._client.write_gatt_char(uuid, encode_int(data))
            _LOGGER.debug("Wrote characteristic %s with value: %s", str(uuid), value)
        except (BleakError, TimeoutError) as e:
            _LOGGER.debug("Failed to write characteristic %s: %s", str(uuid), e)
            raise CommunicationError from e


def decode_int(value: bytearray) -> int:
    """Decode byte-encoded integer value to an integer.

    Parameters
    ----------
    value : bytearray
        Byte-encoded integer value to decode.

    Returns
    -------
    int
        Decoded integer value.

    Notes
    -----
    The byte order is little-endian, and the integer is treated as unsigned.

    """
    return int.from_bytes(value, byteorder="little", signed=False)


def decode_str(value: bytearray) -> str:
    """Decode byte-encoded string to a UTF-8 string.

    Parameters
    ----------
    value : bytearray
        Byte-encoded string to decode.

    Returns
    -------
    str
        Decoded UTF-8 string.

    Raises
    ------
    UnicodeDecodeError
        If the byte sequence cannot be decoded as UTF-8.

    """

    return value.decode("utf-8")


def decode_live_data(value: bytearray) -> LiveDataResponse:
    """Parse bytearray from bulk live data into LiveDataResponse.

    Parameters
    ----------
    value : bytearray
        Byte value from bulk live data characteristic.

    Returns
    -------
    LiveDataResponse
        Object containing 14 decoded values:
        - live_temp: int
        - setpoint_temp: int
        - dc_voltage: float (normalized)
        - handle_temp: float (normalized)
        - pwm_level: int
        - power_src: PowerSource
        - tip_resistance: float (normalized)
        - uptime: float (normalized)
        - movement_time: float (normalized)
        - max_tip_temp_ability: int
        - tip_voltage: int
        - hall_sensor: int
        - operating_mode: OperatingMode
        - estimated_power: float (normalized)

    """
    data = struct.unpack("14I", value)
    return LiveDataResponse(
        data[0],
        data[1],
        data[2] / 10,
        data[3] / 10,
        int(data[4] / 255 * 100),
        PowerSource(data[5]),
        data[6] / 10,
        data[7] / 10,
        data[8] / 10,
        data[9],
        data[10],
        data[11],
        OperatingMode(data[12]),
        data[13] / 10,
    )


def clip(a: int, a_min: int, a_max: int) -> int:
    """Clip a value between max and min.

    Validate if the value is between the defined limits and set
    it to the max/min value if it exceeds one of them.

    Parameters
    ----------
    a : int
        The value to clip
    a_min : int
        The lower bound for the value.
    a_max : int
        The upper bound for the value.

    Returns
    -------
    int
        The clipped value:
        - If `a` is less than `a_min`, return `a_min`.
        - If `a` is greater than `a_max`, return `a_max`.
        - Otherwise, return `a` itself.

    """
    a = min(a, a_max)
    return max(a, a_min)


def encode_int(value: int) -> bytes:
    """Encode integer as byte.

    Parameters
    ----------
    value : int
        An integer value to encode.

    Returns
    -------
    bytes
        Byte representation of the integer value in unsigned little-endian format,
        occupying 2 bytes.

    """
    return value.to_bytes(2, byteorder="little", signed=False)


def encode_lang_code(language_code: str | LanguageCode) -> int:
    """Encode language code to its hashed integer representation.

    ironOS stores languages as hashed integer representations of the language code.
    This method uses the same hash algorithm used in ironOS.

    Parameters
    ----------
    language_code : str | LanguageCode
        The language code as a string or as a member of LanguageCode enum.

    Returns
    -------
    int
        The hashed integer value of language_code.

    Notes
    -----
    If language_code is a member of LanguageCode enum, its integer value is returned directly.
    Otherwise, language_code is hashed using SHA-1, and the resulting hash is converted to an integer
    and returned.

    """
    if isinstance(language_code, LanguageCode):
        return int(language_code.value)

    return int(hashlib.sha1(language_code.encode("utf-8")).hexdigest(), 16) % 0xFFFF


def decode_lang_code(raw: bytearray) -> LanguageCode | int | None:
    """Decode hashed value to language code.

    Parameters
    ----------
    raw : bytearray
        A byte-encoded integer value representing the hashed language code.

    Returns
    -------
    LanguageCode | int | None
        The LanguageCode enum member corresponding to the hashed integer value,
        or the integer value itself if it couldn't be matched to a known LanguageCode.
        Returns None if decoding fails.

    """
    try:
        return LanguageCode(decode_int(raw))
    except ValueError:
        return decode_int(raw)


# Map uuid, decoding, encoding and input sanitizing methods for each characteristic
CHAR_MAP: dict[Characteristic, tuple] = {
    CharBulk.LIVE_DATA: (const.CHAR_UUID_BULK_LIVE_DATA, decode_live_data),
    CharBulk.BUILD: (const.CHAR_UUID_BULK_BUILD, decode_str),
    CharBulk.DEVICE_SN: (
        const.CHAR_UUID_BULK_DEVICE_SN,
        lambda x: f"{decode_int(x):016x}",
    ),
    CharBulk.DEVICE_ID: (
        const.CHAR_UUID_BULK_DEVICE_ID,
        lambda x: f"{decode_int(x):x}",
    ),
    CharLive.LIVE_TEMP: (const.CHAR_UUID_LIVE_LIVE_TEMP, decode_int),
    CharLive.SETPOINT_TEMP: (const.CHAR_UUID_LIVE_SETPOINT_TEMP, decode_int),
    CharLive.DC_VOLTAGE: (
        const.CHAR_UUID_LIVE_DC_VOLTAGE,
        lambda x: decode_int(x) / 10,
    ),
    CharLive.HANDLE_TEMP: (
        const.CHAR_UUID_LIVE_HANDLE_TEMP,
        lambda x: decode_int(x) / 10,
    ),
    CharLive.PWM_LEVEL: (
        const.CHAR_UUID_LIVE_PWM_LEVEL,
        lambda x: int(decode_int(x) / 255 * 100),  # convert to percent
    ),
    CharLive.POWER_SRC: (
        const.CHAR_UUID_LIVE_POWER_SRC,
        lambda x: PowerSource(decode_int(x)),
    ),
    CharLive.TIP_RESISTANCE: (
        const.CHAR_UUID_LIVE_TIP_RESISTANCE,
        lambda x: decode_int(x) / 10,
    ),
    CharLive.UPTIME: (
        const.CHAR_UUID_LIVE_UPTIME,
        lambda x: decode_int(x) / 10,
    ),
    CharLive.MOVEMENT_TIME: (
        const.CHAR_UUID_LIVE_MOVEMENT_TIME,
        lambda x: decode_int(x) / 10,
    ),
    CharLive.TIP_VOLTAGE: (
        const.CHAR_UUID_LIVE_TIP_VOLTAGE,
        decode_int,
    ),
    CharLive.HALL_SENSOR: (const.CHAR_UUID_LIVE_HALL_SENSOR, decode_int),
    CharLive.OPERATING_MODE: (
        const.CHAR_UUID_LIVE_OPERATING_MODE,
        lambda x: OperatingMode(decode_int(x)),
    ),
    CharLive.ESTIMATED_POWER: (const.CHAR_UUID_LIVE_ESTIMATED_POWER, decode_int),
    CharSetting.SETPOINT_TEMP: (
        const.CHAR_UUID_SETTINGS_SETPOINT_TEMP,
        decode_int,
        int,
        lambda x: clip(x, 10, 450),
    ),
    CharSetting.SLEEP_TEMP: (
        const.CHAR_UUID_SETTINGS_SLEEP_TEMP,
        decode_int,
        int,
        lambda x: clip(x, 10, 450),
    ),
    CharSetting.SLEEP_TIMEOUT: (
        const.CHAR_UUID_SETTINGS_SLEEP_TIMEOUT,
        decode_int,
        int,
        lambda x: clip(x, 0, 15),
    ),
    CharSetting.MIN_DC_VOLTAGE_CELLS: (
        const.CHAR_UUID_SETTINGS_MIN_DC_VOLTAGE_CELLS,
        lambda x: BatteryType(decode_int(x)),
        lambda x: x.value if isinstance(x, BatteryType) else int(x),
        lambda x: clip(x, 0, 4),
    ),
    CharSetting.MIN_VOLTAGE_PER_CELL: (
        const.CHAR_UUID_SETTINGS_MIN_VOLTAGE_PER_CELL,
        lambda x: decode_int(x) / 10,
        lambda x: int(x * 10),
        lambda x: clip(x, 24, 38),
    ),
    CharSetting.QC_IDEAL_VOLTAGE: (
        const.CHAR_UUID_SETTINGS_QC_IDEAL_VOLTAGE,
        lambda x: decode_int(x) / 10,
        lambda x: int(x * 10),
        lambda x: clip(x, 90, 220),
    ),
    CharSetting.ORIENTATION_MODE: (
        const.CHAR_UUID_SETTINGS_ORIENTATION_MODE,
        lambda x: ScreenOrientationMode(decode_int(x)),
        lambda x: x.value if isinstance(x, ScreenOrientationMode) else int(x),
        lambda x: clip(x, 0, 2),
    ),
    CharSetting.ACCEL_SENSITIVITY: (
        const.CHAR_UUID_SETTINGS_ACCEL_SENSITIVITY,
        decode_int,
        int,
        lambda x: clip(x, 0, 9),
    ),
    CharSetting.ANIMATION_LOOP: (
        const.CHAR_UUID_SETTINGS_ANIMATION_LOOP,
        decode_int,
        int,
        lambda x: clip(x, 0, 1),
    ),
    CharSetting.ANIMATION_SPEED: (
        const.CHAR_UUID_SETTINGS_ANIMATION_SPEED,
        lambda x: AnimationSpeed(decode_int(x)),
        lambda x: x.value if isinstance(x, AnimationSpeed) else int(x),
        lambda x: clip(x, 0, 3),
    ),
    CharSetting.AUTOSTART_MODE: (
        const.CHAR_UUID_SETTINGS_AUTOSTART_MODE,
        lambda x: AutostartMode(decode_int(x)),
        lambda x: x.value if isinstance(x, AutostartMode) else int(x),
        lambda x: clip(x, 0, 3),
    ),
    CharSetting.SHUTDOWN_TIME: (
        const.CHAR_UUID_SETTINGS_SHUTDOWN_TIME,
        decode_int,
        int,
        lambda x: clip(x, 0, 60),
    ),
    CharSetting.COOLING_TEMP_BLINK: (
        const.CHAR_UUID_SETTINGS_COOLING_TEMP_BLINK,
        decode_int,
        int,
        lambda x: clip(x, 0, 1),
    ),
    CharSetting.IDLE_SCREEN_DETAILS: (
        const.CHAR_UUID_SETTINGS_IDLE_SCREEN_DETAILS,
        decode_int,
        int,
        lambda x: clip(x, 0, 1),
    ),
    CharSetting.SOLDER_SCREEN_DETAILS: (
        const.CHAR_UUID_SETTINGS_SOLDER_SCREEN_DETAILS,
        decode_int,
        int,
        lambda x: clip(x, 0, 1),
    ),
    CharSetting.TEMP_UNIT: (
        const.CHAR_UUID_SETTINGS_TEMP_UNIT,
        lambda x: TempUnit(decode_int(x)),
        lambda x: x.value if isinstance(x, TempUnit) else int(x),
        lambda x: clip(x, 0, 1),
    ),
    CharSetting.DESC_SCROLL_SPEED: (
        const.CHAR_UUID_SETTINGS_DESC_SCROLL_SPEED,
        lambda x: ScrollSpeed(decode_int(x)),
        lambda x: x.value if isinstance(x, ScrollSpeed) else int(x),
        lambda x: clip(x, 0, 1),
    ),
    CharSetting.LOCKING_MODE: (
        const.CHAR_UUID_SETTINGS_LOCKING_MODE,
        lambda x: LockingMode(decode_int(x)),
        lambda x: x.value if isinstance(x, LockingMode) else int(x),
        lambda x: clip(x, 0, 2),
    ),
    CharSetting.KEEP_AWAKE_PULSE_POWER: (
        const.CHAR_UUID_SETTINGS_KEEP_AWAKE_PULSE_POWER,
        lambda x: decode_int(x) / 10,
        lambda x: int(x * 10),
        lambda x: clip(x, 0, 99),
    ),
    CharSetting.KEEP_AWAKE_PULSE_DELAY: (
        const.CHAR_UUID_SETTINGS_KEEP_AWAKE_PULSE_DELAY,
        decode_int,
        int,
        lambda x: clip(x, 0, 9),
    ),
    CharSetting.KEEP_AWAKE_PULSE_DURATION: (
        const.CHAR_UUID_SETTINGS_KEEP_AWAKE_PULSE_DURATION,
        decode_int,
        int,
        lambda x: clip(x, 0, 9),
    ),
    CharSetting.VOLTAGE_DIV: (
        const.CHAR_UUID_SETTINGS_VOLTAGE_DIV,
        decode_int,
        int,
        lambda x: clip(x, 360, 900),
    ),
    CharSetting.BOOST_TEMP: (
        const.CHAR_UUID_SETTINGS_BOOST_TEMP,
        decode_int,
        int,
        lambda x: clip(x, 250, 450) if x != 0 else 0,
    ),
    CharSetting.CALIBRATION_OFFSET: (
        const.CHAR_UUID_SETTINGS_CALIBRATION_OFFSET,
        decode_int,
        int,
        lambda x: clip(x, 100, 2500),
    ),
    CharSetting.POWER_LIMIT: (
        const.CHAR_UUID_SETTINGS_POWER_LIMIT,
        lambda x: decode_int(x) / 10,
        lambda x: int(x * 10),
        lambda x: clip(x, 0, 120),
    ),
    CharSetting.INVERT_BUTTONS: (
        const.CHAR_UUID_SETTINGS_INVERT_BUTTONS,
        decode_int,
        int,
        lambda x: clip(x, 0, 1),
    ),
    CharSetting.TEMP_INCREMENT_LONG: (
        const.CHAR_UUID_SETTINGS_TEMP_INCREMENT_LONG,
        decode_int,
        int,
        lambda x: clip(x, 5, 90),
    ),
    CharSetting.TEMP_INCREMENT_SHORT: (
        const.CHAR_UUID_SETTINGS_TEMP_INCREMENT_SHORT,
        decode_int,
        int,
        lambda x: clip(x, 1, 50),
    ),
    CharSetting.HALL_SENSITIVITY: (
        const.CHAR_UUID_SETTINGS_HALL_SENSITIVITY,
        decode_int,
        int,
        lambda x: clip(x, 0, 9),
    ),
    CharSetting.ACCEL_WARN_COUNTER: (
        const.CHAR_UUID_SETTINGS_ACCEL_WARN_COUNTER,
        decode_int,
        int,
        lambda x: clip(x, 0, 9),
    ),
    CharSetting.PD_WARN_COUNTER: (
        const.CHAR_UUID_SETTINGS_PD_WARN_COUNTER,
        decode_int,
        int,
        lambda x: clip(x, 0, 9),
    ),
    CharSetting.UI_LANGUAGE: (
        const.CHAR_UUID_SETTINGS_UI_LANGUAGE,
        decode_lang_code,
        encode_lang_code,
        lambda x: clip(x, 0, 65535),
    ),
    CharSetting.PD_NEGOTIATION_TIMEOUT: (
        const.CHAR_UUID_SETTINGS_PD_NEGOTIATION_TIMEOUT,
        lambda x,: decode_int(x) / 10,
        lambda x: int(x * 10),
        lambda x: clip(x, 0, 50),
    ),
    CharSetting.DISPLAY_INVERT: (
        const.CHAR_UUID_SETTINGS_DISPLAY_INVERT,
        decode_int,
        int,
        lambda x: clip(x, 0, 1),
    ),
    CharSetting.DISPLAY_BRIGHTNESS: (
        const.CHAR_UUID_SETTINGS_DISPLAY_BRIGHTNESS,
        lambda x: int((decode_int(x) + 24) / 25),  # convert to values 1-5
        lambda x: int(25 * x - 24),
        lambda x: clip(x, 1, 101),
    ),
    CharSetting.LOGO_DURATION: (
        const.CHAR_UUID_SETTINGS_LOGO_DURATION,
        lambda x: LogoDuration(decode_int(x)),
        lambda x: x.value if isinstance(x, LockingMode) else int(x),
        lambda x: clip(x, 0, 6),
    ),
    CharSetting.CALIBRATE_CJC: (
        const.CHAR_UUID_SETTINGS_CALIBRATE_CJC,
        decode_int,
        int,
        lambda x: clip(x, 0, 1),
    ),
    CharSetting.BLE_ENABLED: (
        const.CHAR_UUID_SETTINGS_BLE_ENABLED,
        decode_int,
        int,
        lambda _: 0,
    ),
    CharSetting.USB_PD_MODE: (
        const.CHAR_UUID_SETTINGS_USB_PD_MODE,
        decode_int,
        int,
        lambda x: clip(x, 0, 1),
    ),
    CharSetting.SETTINGS_SAVE: (
        const.CHAR_UUID_SETTINGS_SAVE,
        decode_int,
        int,
        lambda _: 1,
    ),
    CharSetting.SETTINGS_RESET: (
        const.CHAR_UUID_SETTINGS_RESET,
        decode_int,
        int,
        lambda _: 1,
    ),
}



"""Pynecil - Python library to communicate with Pinecil V2 soldering irons via Bluetooth."""

__version__ = "2.0.2"

from .client import Pynecil, discover
from .exceptions import CommunicationError
from .types import (
    AnimationSpeed,
    AutostartMode,
    BatteryType,
    CharBulk,
    CharLive,
    CharSetting,
    DeviceInfoResponse,
    LanguageCode,
    LiveDataResponse,
    LockingMode,
    LogoDuration,
    OperatingMode,
    PowerSource,
    ScreenOrientationMode,
    ScrollSpeed,
    SettingsDataResponse,
    TempUnit,
)

__all__ = [
    "AnimationSpeed",
    "AutostartMode",
    "BatteryType",
    "CharBulk",
    "CharLive",
    "CharSetting",
    "CommunicationError",
    "DeviceInfoResponse",
    "discover",
    "LanguageCode",
    "LiveDataResponse",
    "LockingMode",
    "LogoDuration",
    "OperatingMode",
    "PowerSource",
    "Pynecil",
    "ScreenOrientationMode",
    "ScrollSpeed",
    "SettingsDataResponse",
    "TempUnit",
]



"""Exceptions for Pynecil."""


class CommunicationError(Exception):
    """Exception raised for communication failures."""

    def __init__(self, message="Communication error occurred."):
        """Initialize the CommunicationError exception.

        Parameters
        ----------
        message : str, optional
            Error message describing the communication failure, by default
            "Communication error occurred."

        """
        self.message = message
        super().__init__(self.message)

    def __str__(self):
        """Return a string representation of the exception.

        Returns
        -------
        str
            Formatted string containing the class name and the error message.

        """
        return f"{self.__class__.__name__}: {self.message}"



"""Types for Pynecil."""

from dataclasses import dataclass
from enum import Enum
from typing import TypedDict


class Characteristic(Enum):
    """Base class for Characteristics."""


class CharLive(Characteristic, Enum):
    """Characteristics from live service."""

    LIVE_TEMP = 0
    SETPOINT_TEMP = 1
    DC_VOLTAGE = 2
    HANDLE_TEMP = 3
    PWM_LEVEL = 4
    POWER_SRC = 5
    TIP_RESISTANCE = 6
    UPTIME = 7
    MOVEMENT_TIME = 8
    MAX_TIP_TEMP_ABILITY = 9
    TIP_VOLTAGE = 10
    HALL_SENSOR = 11
    OPERATING_MODE = 12
    ESTIMATED_POWER = 13


class CharSetting(Characteristic, Enum):
    """Characteristics from settings service."""

    SETPOINT_TEMP = 0
    SLEEP_TEMP = 1
    SLEEP_TIMEOUT = 2
    MIN_DC_VOLTAGE_CELLS = 3
    MIN_VOLTAGE_PER_CELL = 4
    QC_IDEAL_VOLTAGE = 5
    ORIENTATION_MODE = 6
    ACCEL_SENSITIVITY = 7
    ANIMATION_LOOP = 8
    ANIMATION_SPEED = 9
    AUTOSTART_MODE = 10
    SHUTDOWN_TIME = 11
    COOLING_TEMP_BLINK = 12
    IDLE_SCREEN_DETAILS = 13
    SOLDER_SCREEN_DETAILS = 14
    TEMP_UNIT = 15
    DESC_SCROLL_SPEED = 16
    LOCKING_MODE = 17
    KEEP_AWAKE_PULSE_POWER = 18
    KEEP_AWAKE_PULSE_DELAY = 19
    KEEP_AWAKE_PULSE_DURATION = 20
    VOLTAGE_DIV = 21
    BOOST_TEMP = 22
    CALIBRATION_OFFSET = 23
    POWER_LIMIT = 24
    INVERT_BUTTONS = 25
    TEMP_INCREMENT_LONG = 26
    TEMP_INCREMENT_SHORT = 27
    HALL_SENSITIVITY = 28
    ACCEL_WARN_COUNTER = 29
    PD_WARN_COUNTER = 30
    UI_LANGUAGE = 31
    PD_NEGOTIATION_TIMEOUT = 32
    DISPLAY_INVERT = 33
    DISPLAY_BRIGHTNESS = 34
    LOGO_DURATION = 35
    CALIBRATE_CJC = 36
    BLE_ENABLED = 37
    USB_PD_MODE = 38
    SETTINGS_RESET = 98
    SETTINGS_SAVE = 99


class CharBulk(Characteristic, Enum):
    """Characteristics from bulk service."""

    LIVE_DATA = 0
    ACCEL_NAME = 1
    BUILD = 2
    DEVICE_SN = 3
    DEVICE_ID = 4


class PowerSource(Enum):
    """Power source types."""

    DC = 0
    QC = 1
    PD_VBUS = 2
    PD = 3


class OperatingMode(Enum):
    """Operating modes."""

    IDLE = 0
    SOLDERING = 1
    BOOST = 2
    SLEEPING = 3
    SETTINGS = 4
    DEBUG = 5


class LanguageCode(Enum):
    """Language Codes."""

    BE = 60301
    BG = 15395
    CS = 36791
    DA = 63942
    DE = 5496
    EL = 5003
    EN = 41431
    ES = 38713
    ET = 18074
    FI = 25411
    FR = 38783
    HR = 49773
    HU = 19902
    IT = 57867
    JA_JP = 2385
    LT = 5183
    NB = 31043
    NL = 22266
    NL_BE = 55046
    PL = 55968
    PT = 56922
    RO = 61480
    RU = 26979
    SK = 13916
    SL = 21931
    SR_CYRL = 41427
    SR_LATN = 61017
    SV = 65456
    TR = 9120
    UK = 29374
    VI = 20758
    YUE_HK = 17119
    ZH_CN = 44731
    ZH_TW = 34289


class BatteryType(Enum):
    """Battery type (cell count)."""

    NO_BATTERY = 0  # DC power supply
    BATTERY_3S = 1
    BATTERY_4S = 2
    BATTERY_5S = 3
    BATTERY_6S = 4


class ScreenOrientationMode(Enum):
    """Screen orientation mode."""

    RIGHT_HANDED = 0
    LEFT_HANDED = 1
    AUTO = 2


class AnimationSpeed(Enum):
    """Animation speed."""

    OFF = 0
    SLOW = 1
    MEDIUM = 2
    FAST = 3


class AutostartMode(Enum):
    """Autostart mode."""

    DISABLED = 0
    SOLDERING = 1
    SLEEPING = 2
    IDLE = 3


class TempUnit(Enum):
    """Temperature unit."""

    CELSIUS = 0
    FAHRENHEIT = 1


class ScrollSpeed(Enum):
    """Description scroll speed."""

    SLOW = 0
    FAST = 1


class LockingMode(Enum):
    """Locking mode."""

    OFF = 0
    BOOST_ONLY = 1
    FULL_LOCKING = 2


class LogoDuration(Enum):
    """Logo duration."""

    OFF = 0
    SECONDS_1 = 1
    SECONDS_2 = 2
    SECONDS_3 = 3
    SECONDS_4 = 4
    SECONDS_5 = 5
    LOOP = 6


@dataclass
class DeviceInfoResponse:
    """Response data with information about the Pinecil device.

    Attributes
    ----------
    build: str | None
        ironOS build version of the device (e.g. 2.22)
    device_id:  str | None
        Identifier of the device
    address: str | None
        Bluetooth address of the device
    device_sn: str | None
        Serial number of the device
    name: str | None
        Name of the device

    """

    build: str | None = None
    device_id: str | None = None
    address: str | None = None
    device_sn: str | None = None
    name: str | None = None
    is_synced: bool = False


@dataclass
class LiveDataResponse:
    """Live data response class.

    Attributes
    ----------
    live_temp: int | None
        Temperature of the tip (in °C)
    setpoint_temp: int | None
        Setpoint temperature (in °C)
    dc_voltage: float | None
        DC input voltage
    handle_temp: float | None
        Handle temperature (in °C)
    pwm_level: int | None
        Power level (0-100%)
    power_src: PowerSource | None
        Power source (e.g. USB-PD/QC or DC)
    tip_resistance: float | None
        Resistance of the tip (in Ω)
    uptime: float | None
        Uptime of the device (in seconds)
    movement_time: float | None
        Last movement time (in seconds)
    max_tip_temp_ability: int | None
        Maximum temperature supported by the tip (in °C)
    tip_voltage: int | None
        Raw tip voltage (in μV)
    hall_sensor: int | None
        Hall effect strength (if hall sensor is installed)
    operating_mode: OperatingMode | None
        Current operating mode of the device (e.g. soldering, idle...)
    estimated_power: float | None
        Estimated power usage (in Watt)

    """

    live_temp: int | None = None
    setpoint_temp: int | None = None
    dc_voltage: float | None = None
    handle_temp: float | None = None
    pwm_level: int | None = None
    power_src: PowerSource | None = None
    tip_resistance: float | None = None
    uptime: float | None = None
    movement_time: float | None = None
    max_tip_temp_ability: int | None = None
    tip_voltage: int | None = None
    hall_sensor: int | None = None
    operating_mode: OperatingMode | None = None
    estimated_power: float | None = None


class SettingsDataResponse(TypedDict, total=False):
    """Settings data response class.

    Attributes
    ----------
    setpoint_temp: int | None
        Setpoint temperature (in °C, 10-450)
    sleep_temp: int | None
        Temperature to drop to in sleep mode (in °C, 10-450)
    sleep_timeout: int | None
        timeout till sleep mode (in minutes , 0-15)
    min_dc_voltage_cells: BatteryType | None
        Voltage to cut out at for under voltage when powered by DC jack (0=DC, 1=3S, 2=4S, 3=5S, 4=6S)
    min_voltage_per_cell: float | None
        Minimum allowed voltage per cell (in V, 2.4-3.8, step=0.1)
    qc_ideal_voltage: float | None
        QC3.0 maximum voltage (9.0-22.0V, step=0.1)
    orientation_mode: ScreenOrientationMode | None
        Screen orientation (right-handed, left-handed, Auto)
    accel_sensitivity: int | None
        Motion sensitivity (0-9)
    animation_loop: int | None
        Animation loop switch (0=off, 1=on)
    animation_speed: AnimationSpeed | None
        Animation speed (off, slow, medium, fast)
    autostart_mode: AutostartMode | None
        Mode to enter on start-up (disabled, soldering, sleeping, idle)
    shutdown_time: int | None
        Time until unit shuts down if not moved (in seconds, 0-60)
    cooling_temp_blink: int | None
        Blink temperature on the cool down screen until its <50C (0=off, 1=on)
    idle_screen_details: int | None
        Show details on idle screen (0=off, 1=on)
    solder_screen_details: int | None
        Show details on soldering screen (0=off, 1=on)
    temp_unit: TempUnit | None
        Temperature unit (0=Celsius, 1=Fahrenheit)
    desc_scroll_speed: ScrollSpeed | None
        Description scroll speed (0=slow, 1=fast)
    locking_mode: LockingMode | None
        Allow locking buttons (off, boost mode only, full locking)
    keep_awake_pulse_power: float | None
        Intensity of power of keep-awake-pulse in W (0-9.9)
    keep_awake_pulse_delay: int | None
        Delay before keep-awake-pulse is triggered (1-9 x 2.5s)
    keep_awake_pulse_duration: int | None
        Keep-awake-pulse duration (1-9 x 250ms)
    voltage_div: int | None
        Voltage divisor factor (360-900)
    boost_temp: int | None
        Boost mode set point temperature (in °C, 0-450)
    calibration_offset: int | None
        Calibration offset for the installed tip (in µV, 100-2500)
    power_limit: float | None
        Maximum power allowed to output (in W, 0-12W, step=0.1)
    invert_buttons: int | None
        Change the plus and minus button assigment (0=off, 1=invert)
    temp_increment_long: int | None
        Temperature-change-increment on long button press in degree (5-90)
    temp_increment_short: int | None
        Temperature-change-increment on short button press in degree (1-50)
    hall_sensitivity: int | None
        Hall effect sensor sensitivity (0-9)
    accel_warn_counter: int | None
        Warning counter when accelerometer could not be detected (0-9)
    pd_warn_counter: int | None
        Warning counter when PD interface could not be detected (0-9)
    ui_language: LanguageCode | None
        Hashed integer of language code
    pd_negotiation_timeout: float | None
        Power delivery negotiation timeout in seconds (0-5.0, step=0.1)
    display_invert: int | None
        Invert colors of display (0=off, 1=on)
    display_brightness: int | None
        Display brightness (1-5)
    logo_duration: LogoDuration | None
        Boot logo duration (off, 1-5seconds, loop ∞)
    calibrate_cjc: int | None
        Enable CJC calibration at next boot
    ble_enabled: int | None
        Disable BLE
    usb_pd_mode: int | None
        PPS & EPR USB Power delivery mode (0=off, 1=on)
    settings_save: int | None
        Save settings to flash
    settings_reset: int | None
        Reset settings to default values

    """

    setpoint_temp: int | None
    sleep_temp: int | None
    sleep_timeout: int | None
    min_dc_voltage_cells: BatteryType | None
    min_voltage_per_cell: float | None
    qc_ideal_voltage: float | None
    orientation_mode: ScreenOrientationMode | None
    accel_sensitivity: int | None
    animation_loop: int | None
    animation_speed: AnimationSpeed | None
    autostart_mode: AutostartMode | None
    shutdown_time: int | None
    cooling_temp_blink: int | None
    idle_screen_details: int | None
    solder_screen_details: int | None
    temp_unit: TempUnit | None
    desc_scroll_speed: ScrollSpeed | None
    locking_mode: LockingMode | None
    keep_awake_pulse_power: float | None
    keep_awake_pulse_delay: int | None
    keep_awake_pulse_duration: int | None
    voltage_div: int | None
    boost_temp: int | None
    calibration_offset: int | None
    power_limit: float | None
    invert_buttons: int | None
    temp_increment_long: int | None
    temp_increment_short: int | None
    hall_sensitivity: int | None
    accel_warn_counter: int | None
    pd_warn_counter: int | None
    ui_language: LanguageCode | None
    pd_negotiation_timeout: float | None
    display_invert: int | None
    display_brightness: int | None
    logo_duration: LogoDuration | None
    calibrate_cjc: int | None
    ble_enabled: int | None
    usb_pd_mode: int | None
    settings_save: int | None
    settings_reset: int | None



"""Constants for Pynecil."""

from uuid import UUID

SVC_UUID_BULK = UUID("9eae1000-9d0d-48c5-aa55-33e27f9bc533")
CHAR_UUID_BULK_LIVE_DATA = UUID("9eae1001-9d0d-48c5-aa55-33e27f9bc533")
CHAR_UUID_BULK_ACCEL_NAME = UUID("9eae1002-9d0d-48c5-aa55-33e27f9bc533")
CHAR_UUID_BULK_BUILD = UUID("9eae1003-9d0d-48c5-aa55-33e27f9bc533")
CHAR_UUID_BULK_DEVICE_SN = UUID("9eae1004-9d0d-48c5-aa55-33e27f9bc533")
CHAR_UUID_BULK_DEVICE_ID = UUID("9eae1005-9d0d-48c5-aa55-33e27f9bc533")

SVC_UUID_LIVE = UUID("d85ef000-168e-4a71-aa55-33e27f9bc533")
CHAR_UUID_LIVE_LIVE_TEMP = UUID("d85ef001-168e-4a71-aa55-33e27f9bc533")
CHAR_UUID_LIVE_SETPOINT_TEMP = UUID("d85ef002-168e-4a71-aa55-33e27f9bc533")
CHAR_UUID_LIVE_DC_VOLTAGE = UUID("d85ef003-168e-4a71-aa55-33e27f9bc533")
CHAR_UUID_LIVE_HANDLE_TEMP = UUID("d85ef004-168e-4a71-aa55-33e27f9bc533")
CHAR_UUID_LIVE_PWM_LEVEL = UUID("d85ef005-168e-4a71-aa55-33e27f9bc533")
CHAR_UUID_LIVE_POWER_SRC = UUID("d85ef006-168e-4a71-aa55-33e27f9bc533")
CHAR_UUID_LIVE_TIP_RESISTANCE = UUID("d85ef007-168e-4a71-aa55-33e27f9bc533")
CHAR_UUID_LIVE_UPTIME = UUID("d85ef008-168e-4a71-aa55-33e27f9bc533")
CHAR_UUID_LIVE_MOVEMENT_TIME = UUID("d85ef009-168e-4a71-aa55-33e27f9bc533")
CHAR_UUID_LIVE_MAX_TIP_TEMP_ABILITY = UUID("d85ef00a-168e-4a71-aa55-33e27f9bc533")
CHAR_UUID_LIVE_TIP_VOLTAGE = UUID("d85ef00b-168e-4a71-aa55-33e27f9bc533")
CHAR_UUID_LIVE_HALL_SENSOR = UUID("d85ef00c-168e-4a71-aa55-33e27f9bc533")
CHAR_UUID_LIVE_OPERATING_MODE = UUID("d85ef00d-168e-4a71-aa55-33e27f9bc533")
CHAR_UUID_LIVE_ESTIMATED_POWER = UUID("d85ef00e-168e-4a71-aa55-33e27f9bc533")

SVC_UUID_SETTINGS = UUID("f6d80000-5a10-4eba-aa55-33e27f9bc533")
CHAR_UUID_SETTINGS_SAVE = UUID("f6d7ffff-5a10-4eba-aa55-33e27f9bc533")
CHAR_UUID_SETTINGS_RESET = UUID("f6d7fffe-5a10-4eba-aa55-33e27f9bc533")
CHAR_UUID_SETTINGS_SETPOINT_TEMP = UUID("f6d70000-5a10-4eba-aa55-33e27f9bc533")
CHAR_UUID_SETTINGS_SLEEP_TEMP = UUID("f6d70001-5a10-4eba-aa55-33e27f9bc533")
CHAR_UUID_SETTINGS_SLEEP_TIMEOUT = UUID("f6d70002-5a10-4eba-aa55-33e27f9bc533")
CHAR_UUID_SETTINGS_MIN_DC_VOLTAGE_CELLS = UUID("f6d70003-5a10-4eba-aa55-33e27f9bc533")
CHAR_UUID_SETTINGS_MIN_VOLTAGE_PER_CELL = UUID("f6d70004-5a10-4eba-aa55-33e27f9bc533")
CHAR_UUID_SETTINGS_QC_IDEAL_VOLTAGE = UUID("f6d70005-5a10-4eba-aa55-33e27f9bc533")
CHAR_UUID_SETTINGS_ORIENTATION_MODE = UUID("f6d70006-5a10-4eba-aa55-33e27f9bc533")
CHAR_UUID_SETTINGS_ACCEL_SENSITIVITY = UUID("f6d70007-5a10-4eba-aa55-33e27f9bc533")
CHAR_UUID_SETTINGS_ANIMATION_LOOP = UUID("f6d70008-5a10-4eba-aa55-33e27f9bc533")
CHAR_UUID_SETTINGS_ANIMATION_SPEED = UUID("f6d70009-5a10-4eba-aa55-33e27f9bc533")
CHAR_UUID_SETTINGS_AUTOSTART_MODE = UUID("f6d7000a-5a10-4eba-aa55-33e27f9bc533")
CHAR_UUID_SETTINGS_SHUTDOWN_TIME = UUID("f6d7000b-5a10-4eba-aa55-33e27f9bc533")
CHAR_UUID_SETTINGS_COOLING_TEMP_BLINK = UUID("f6d7000c-5a10-4eba-aa55-33e27f9bc533")
CHAR_UUID_SETTINGS_IDLE_SCREEN_DETAILS = UUID("f6d7000d-5a10-4eba-aa55-33e27f9bc533")
CHAR_UUID_SETTINGS_SOLDER_SCREEN_DETAILS = UUID("f6d7000e-5a10-4eba-aa55-33e27f9bc533")
CHAR_UUID_SETTINGS_TEMP_UNIT = UUID("f6d7000f-5a10-4eba-aa55-33e27f9bc533")
CHAR_UUID_SETTINGS_DESC_SCROLL_SPEED = UUID("f6d70010-5a10-4eba-aa55-33e27f9bc533")
CHAR_UUID_SETTINGS_LOCKING_MODE = UUID("f6d70011-5a10-4eba-aa55-33e27f9bc533")
CHAR_UUID_SETTINGS_KEEP_AWAKE_PULSE_POWER = UUID("f6d70012-5a10-4eba-aa55-33e27f9bc533")
CHAR_UUID_SETTINGS_KEEP_AWAKE_PULSE_DELAY = UUID("f6d70013-5a10-4eba-aa55-33e27f9bc533")
CHAR_UUID_SETTINGS_KEEP_AWAKE_PULSE_DURATION = UUID(
    "f6d70014-5a10-4eba-aa55-33e27f9bc533"
)
CHAR_UUID_SETTINGS_VOLTAGE_DIV = UUID("f6d70015-5a10-4eba-aa55-33e27f9bc533")
CHAR_UUID_SETTINGS_BOOST_TEMP = UUID("f6d70016-5a10-4eba-aa55-33e27f9bc533")
CHAR_UUID_SETTINGS_CALIBRATION_OFFSET = UUID("f6d70017-5a10-4eba-aa55-33e27f9bc533")
CHAR_UUID_SETTINGS_POWER_LIMIT = UUID("f6d70018-5a10-4eba-aa55-33e27f9bc533")
CHAR_UUID_SETTINGS_INVERT_BUTTONS = UUID("f6d70019-5a10-4eba-aa55-33e27f9bc533")
CHAR_UUID_SETTINGS_TEMP_INCREMENT_LONG = UUID("f6d7001a-5a10-4eba-aa55-33e27f9bc533")
CHAR_UUID_SETTINGS_TEMP_INCREMENT_SHORT = UUID("f6d7001b-5a10-4eba-aa55-33e27f9bc533")
CHAR_UUID_SETTINGS_HALL_SENSITIVITY = UUID("f6d7001c-5a10-4eba-aa55-33e27f9bc533")
CHAR_UUID_SETTINGS_ACCEL_WARN_COUNTER = UUID("f6d7001d-5a10-4eba-aa55-33e27f9bc533")
CHAR_UUID_SETTINGS_PD_WARN_COUNTER = UUID("f6d7001e-5a10-4eba-aa55-33e27f9bc533")
CHAR_UUID_SETTINGS_UI_LANGUAGE = UUID("f6d7001f-5a10-4eba-aa55-33e27f9bc533")
CHAR_UUID_SETTINGS_PD_NEGOTIATION_TIMEOUT = UUID("f6d70020-5a10-4eba-aa55-33e27f9bc533")
CHAR_UUID_SETTINGS_DISPLAY_INVERT = UUID("f6d70021-5a10-4eba-aa55-33e27f9bc533")
CHAR_UUID_SETTINGS_DISPLAY_BRIGHTNESS = UUID("f6d70022-5a10-4eba-aa55-33e27f9bc533")
CHAR_UUID_SETTINGS_LOGO_DURATION = UUID("f6d70023-5a10-4eba-aa55-33e27f9bc533")
CHAR_UUID_SETTINGS_CALIBRATE_CJC = UUID("f6d70024-5a10-4eba-aa55-33e27f9bc533")
CHAR_UUID_SETTINGS_BLE_ENABLED = UUID("f6d70025-5a10-4eba-aa55-33e27f9bc533")
CHAR_UUID_SETTINGS_USB_PD_MODE = UUID("f6d70026-5a10-4eba-aa55-33e27f9bc533")
